import pexpect, sys
from datetime import *
import pdb

class HorizonsDataSource(object):
    HORIZONS_ADDRESS = 'horizons.jpl.nasa.gov'
    HORIZONS_PORT = 6775
    
    def __init__(self, debug=False):
        self.debug = debug
        self.p = pexpect.spawn('telnet')
        self.p.timeout = 10
        self.expect('telnet>')
        self.send('environ define TERM no-resize-vt100') # disable resizing, it caused a long delay and error
        self.expect('telnet>')
        self.send('open %s %d' % (self.HORIZONS_ADDRESS, self.HORIZONS_PORT))
        self._connect()
        
    def send(self, s):
        if self.debug:
            self._print('Sending: %s' % s)
        return self.p.sendline(s) # number of bytes written
        
    def expect(self, s):
        if self.debug:
            self._print('Expecting: %s' % str(s))
        result = self.p.expect(s)
        #if self.debug:
        #    self._print('Match: %s' % self.p.match.string)
        return result

    def _print(self, s, newline=True):
        s = str(s)
        if newline:
            s += "\n"
        sys.stdout.write(s)
        sys.stdout.flush()
        
    def _connect(self):
        self._print('Connecting to Horizons...')
        self.expect('Horizons>')
        self._print('Connected')
        self.connected = True
        self.send('tty 10000 10000') # set huge screen so we never have to scroll
        self.expect('screen reset \(rows,cols\) =')
    
    def get_angles(self, start, end, step, location):
        if not self.connected:
            raise Exception("Not connected to Horizons")
            
        try:
            self.send('sun')
            self.expect('Select \.\.\. \[E\]phemeris')
            self.send('e') # ephemeris
            self.expect('Observe, Elements, Vectors')
            self.send('o') # observe
            i = self.expect(['Coordinate center', 'Use previous center'])
            if i == 1: # coordinates already specified, but the location may have changed. just resend them
                self.send('n')
                self.expect('Coordinate center')
            self.send('coord') # specify long & latt
            self.expect('Cylindrical or Geodetic input')
            self.send('g')
            self.expect('Specify geodetic')
            # long, lat, height(km)
            self.send('%d %d 0' % (location[1], location[0]))
            self.expect('Starting UT')
            self.send(start.strftime("%Y %m %d %H:%M:%S"))
            self.expect('Ending')
            self.send(end.strftime("%Y %m %d %H:%M:%S"))
            self.expect('Output interval')
            self.send(step)
            self.expect('Accept default output')
            self.send('y')
            i = self.expect(['Select table quantities', 'Select\.\.\. \[A\]gain'])
            if i == 0:
                # 2: RA & DEC, 4: AZ & EL
                self.send('4')
            else:
                self.send('r')
            
        
            self.expect('\$\$SOE\s*') # consume up to start of data rows
            self.expect('\$\$EOE') # end of data
            result = self.p.before # string of all rows of data

            self._print('Data received')
            data = self.process_results(result)
        
            self.send('n') # start over from the Horizons> prompt
            self.expect('Horizons>')
            
            return data
        except Exception, e:
            print "Error in Horizons connection: %s" % str(e)
            return []
    
    def process_results(self, result):
        "takes a string of result data from Horizons and turns it into a Python dict with Python values"
        datarows = result.splitlines()
        
        def get_data(r): # create a data tuple from a row of Horizon data
            tokens = r.split()
            if self.debug:
                print repr(tokens)
            az = float(tokens[-2])
            el = float(tokens[-1])
            date_time = "%s %s" % (tokens[0], tokens[1])
            time = datetime.strptime(date_time, "%Y-%b-%d %H:%M:%S")
            return {
                'datetime': time,
                'az': az,
                'el': el
            }
            
        return [get_data(r) for r in datarows]
        
    def close(self):
        self.send('q') # try to close telnet session
        self.p.close()
        self.connected = False

# this is where our local model code will go
class LocalDataSource(object):
    pass
    
class SolarPositionModel(object):
    _location = None
    _utc_offset = None
    _DEFAULT_LOCATION = (47, 238)
    
    def __init__(self, datasource, debug=False):
        if datasource == 'horizons':
            self.datasource = HorizonsDataSource(debug=debug)
        #elif datasource == 'local':
        #    self.datasource = LocalDataSource()
        else:
            raise ValueError("datasource must be 'horizons'")
        
    def get_angles(self, start, end, step):
        
        if self._location == None:
            print "No location specified.  Defaulting to (lat, long) = (%d, %d)" % self._DEFAULT_LOCATION
            self._location = self._DEFAULT_LOCATION
            
        if self._utc_offset == None:
            print "No UTC offset specified.  Defaulting to 0 (UTC)"
            self._utc_offset = 0
        
        start -= timedelta(hours=self._utc_offset)
        end -= timedelta(hours=self._utc_offset)
        
        return self.datasource.get_angles(start, end, step, self._location)
        
    def _adjust_date(self, d):
        "adjusts hour of given date to UTC based on current UTC offset"
        d -= timedelta(hours=self._utc_offset)
        
    def set_location(self, lat, long):
        self._location = (lat, long)
        
    def set_utc_offset(self, offset):
        self._utc_offset = offset

if __name__ == "__main__":
    spm = SolarPositionModel('horizons')
    #spm.set_location(47, 122)
    spm.set_utc_offset(-7)
    
    start = datetime(2010, 5, 26, 11, 00, 1) # y m d h min s
    end = datetime(2010, 5, 26, 13, 20, 0)
    for i in xrange(1,4):
        data = spm.get_angles(start, end, '%i0m' % i) # start, end, step
        for d in data:
            print d['datetime'], ": ",  d['el']